#include <stdio.h>
#include "proms.h"

#define B2C(v,b) (((v)&(1<<b)) ? '1' : '0')

// DDS     = double density bit in configuration byte
// FORMAT  = format bit in configuration byte
// XFORMAT = bit 7 in configuration byte

static const char *u42inputs[8] =
  {"CMD0", "CMD1", "CMD2", "BUSY", "COMPARE", "INDX", "XFORMAT", "CRCER"};

static const char *u43inputs[8] =
  {"U43-0", "FDDAM", "TRK0", "CURTRK<", "TRKCOMP", "DCOMP", "DDS", "FORMAT"};

static const char *u44inputs[8] =
  {"SIDE1", "SEC1", "SEC2", "SEC3", "SEC4", "UNCMP", "STIMER", "TIMER"};


struct {
  char *name;
  byte  cmd;
} commands[] =
  { {"READ",       0x03},
    {"READ_CRC",   0x07},
    {"WRITE",      0x05},
    {"WRITE_DDAM", 0x0F},
    {"SEEK",       0x09},
    {"SEEK0",      0x0D},
    {NULL, 0}
  };


const char *bins(byte b)
{
  static char s[9];
  for(int i=0; i<8; i++) { s[i] = (b&0x80) ? '1' : '0'; b<<=1; }
  return s;
}


const char *getcode(byte codesel)
{
  static char s[20];

  byte bcode = U57[codesel];
  byte dcode = U47[codesel];
  byte useDC = (U56[codesel] & 0x04)==0;
  byte useSR = (U56[codesel] & 0x08)==0;
  byte useTR = (U56[codesel] & 0x10)==0;

  for(int j=0; j<8; j++)
    {
      s[2*j] = (bcode & 0x80) ? '1' : '0';
      if( useDC ) s[2*j+1] = (dcode & 0x80) ? '1' : '0';
      else if( useSR ) s[2*j+1] = 'S';
      else if( useTR ) s[2*j+1] = 'T';
      else             s[2*j+1] = 'X';
      bcode <<= 1;
      dcode <<= 1;
    }
      
  s[16] = 0;
  return s;
}

int main(int argc, char **argv)
{
  for(int i=0; commands[i].name!=NULL; i++)
    {
      byte d = U119[((commands[i].cmd))/2];
      printf("%02X / %02X %c%c%c : %s\n", commands[i].cmd, d,
             (d & 0x20) ? '1' : '0',
             (d & 0x40) ? '1' : '0',
             (d & 0x80) ? '1' : '0',
             commands[i].name);
    }

  for(int i=0; i<0x18; i++) printf("CODESEL %02x: %s\n", i, getcode(i));

  for(int i=0; i<512; i++)
    {
      byte u72 = U72[i];
      byte u73 = U73[i];

      printf("%04X: %s %02X:", i, bins(u72), u73); 
      
      if( u72==0 && u73==0 )
        { printf(" NOP\n"); continue; }

      if( (u72 & 0xC0)==0xC0 )
        {
          // set timer
          int t = (u73 & 0x7F); // ROM0-6
          t |= 0x2480; // always high
          if( u72 & 0x08 ) t |= 0x0100; // ROM11
          if( u73 & 0x80 ) t |= 0x0200; // ROM7
          if( u72 & 0x01 ) t |= 0x0800; // ROM8
          if( u72 & 0x02 ) t |= 0x1000; // ROM9
          if( u72 & 0x04 ) t |= 0x4000; // ROM10

          printf(" SET TIMER %i ticks", 0x8000-t);
        }
      else if( (u72 & 0x88)==0x88 )
        {
          // set short-term timer
          int t = (u73 & 0x0F); // ROM0-3
          printf(" SET STIMER %i ticks", 0x10-t);
        }

      if( (u72 & 0x90)==0x90 )
        {
          //printf(" OUT A, %02X", u73);
          printf(" OUT WEN=%c, LDRBUF=%c, INCWBUF=%c, CRCEN=%c, HDLD=%c, STEPIN=%c, STEP=%c, WGATE=%c",
                 B2C(u73, 7), B2C(u73, 6), B2C(u73, 5), B2C(u73, 4), B2C(u73, 3), B2C(u73, 2), B2C(u73, 1), B2C(u73, 0));
        }

      if( (u72 & 0xA0)==0xA0 )
        {
          static const char *reg[4] = {"CRC", "WDR", "CD0", "CD1"};

          int d = u73 | ((u72&3)<<8);
          //printf(" OUT B, %03X", d);
          byte codesel = (u73&0x0F) | ((u72&0x02)<<3);
          printf(" OUT CODE=%s, REGSEL=%s, LDRTRK=%c, NENDDMTEST=%c, CRCTEST=%c, INCSEC=%c, DONE=%c, CLKSEL=%c, NLDWOUTWD=%c",
                 getcode(codesel), reg[U56[codesel] & 0x03], B2C(U56[codesel], 6), B2C(U56[codesel], 5), 
                 B2C(u73,7), B2C(u73,6), B2C(u73,5), B2C(u73,4), B2C(u72,0));
        }

      if( (u72 & 0x80)==0x00 )
        {
          // conditional jump
          //byte cond = ((u72&0x01) ? 0x04 : 0) | (u72&0x02) | ((u72&0x04) ? 0x01 : 0);
          byte cond = (u72&0x07);

          // C=COND, B=O6, A=O7
          // 0       0     0       | 0
          // 0       0     1       | 1
          // 0       1     0       | 1
          // 0       1     1       | 1
          // 1       0     0       | 1
          // 1       0     1       | 1
          // 1       1     0       | 0
          // 1       1     1       | 1
          // jump if COND=0 and O7=0 and O6=0
          //         COND=1 and O7=0 and O6=1
          
          printf(" JMP %04X if ", u73 | ((u72 & 0x20) ? 0x100 : 0));
          if( (u72 & 0x40)==0 ) printf("~");

          int U42en =  (u72 & 0x08) && !(u72 & 0x10);
          int U43en = !(u72 & 0x08) &&  (u72 & 0x10);
          int U44en =  (u72 & 0x08) &&  (u72 & 0x10);

          if( U42en ) printf("%s ", u42inputs[cond]);
          if( U43en ) printf("%s ", u43inputs[cond]);
          if( U44en ) printf("%s ", u44inputs[cond]);
        }

      printf("\n");
    }
  
  return 0;
}
